/**
 * @file MotorController.cpp
 * @brief ステッピングモータ制御
 * @author Naoki Okayama
 * @date 2017/05/31
 */
#include <functional>
#include <iostream>
#include <wiringPi.h>
#include "PortDefine.h"
#include "CSwitchPhy.h"
#include "MotorController.h"

/*! @brief
 * CLOCK = 19.2MHz / (RPS * 200(モータステップ数) * PWMRANGE)
 * CLOCK = RPS_TO_CLOCK / RPS
 * とする。
 * RPS_TO_CLOCK = 93.75となるが、小数となる。
 * 使用時にintへのキャストを忘れないこと
 */
#define RPS_TO_CLOCK ((float)93.75)

#define SLEEP P_MO_SLEEP                  /*!< SLEEP端子のGPIO番号 */
#define DIR P_MO_DIR                      /*!< DIR端子のGPIO番号 */
#define STEP P_MO_STEP                    /*!< STEP端子のGPIO番号 */

#define DIR_CW 1                          /*!< CWの定義 */
#define DIR_CCW 0                         /*!< CCWの定義 */

#define ANGLE_BY_STEP 1.8                 /*!< 1STEPあたりの回転角度[deg] */

#define ANTI_WINDUP 1                     /*!< アンチワインドアップ 1:ON 0:OFF */
#define I_CLAUSE_LIMIT  ((float)400.0)    /*!< I項リミット値 */
#if (TYPE==A)
#define DEG_TO_MM ((float)2.5/4/360.0)              /*!< 回転角度からMmへ変換する係数 */
#elif (TYPE==B)
#define DEG_TO_MM ((float)1)
#else

#endif
#define P_GAIN ((float)0.5)             /*!< Pゲイン */
#define I_GAIN ((float)0.75)             /*!< Iゲイン */

static MotorController *gThis;

void intPWMEdge(void){
  int dir;
  int sleep;

  dir = digitalRead(DIR);
  sleep = digitalRead(SLEEP);

  if(sleep != 0){
    if(dir == DIR_CW){
      gThis->angle -= ANGLE_BY_STEP;
    }else if(dir == DIR_CCW){
      gThis->angle += ANGLE_BY_STEP;
    }else{
      /* 回転方向不明の場合は処理なし */
    }
  }
}


//////////////////////////////////////////////////////////
// public
//////////////////////////////////////////////////////////
/**
 * @brief コンストラクタ
 * @param -
 * @return -
 * @sa -
 */
MotorController::MotorController(void){
  gThis = this;
}

/**
 * @brief デストラクタ
 * @param -
 * @return -
 * @sa -
 */
MotorController::~MotorController(void){
  if(this->gpioEnabled == true){
    pinMode(SLEEP, OUTPUT);
    pinMode(DIR ,OUTPUT);
    pinMode(STEP, OUTPUT);

    digitalWrite(SLEEP, 0);
    digitalWrite(DIR, 0);
    digitalWrite(STEP, 0);
  }
}


/**
 * @brief モータコントローラの初期化を行う
 * @param -
 * @return 状態（成功：RET_SUCCESS 失敗：RET_FAILED）
 * @sa -
 */
int MotorController::initMotorController(bool bInitGpio){
  this->iClause = 0.0;
  this->stopFlag = true;

  if(true == bInitGpio){
    if(wiringPiSetup() == -1){
      this->gpioEnabled = false;

      return RET_FAILED;
    }else{
      this->gpioEnabled = true;
    }
  }

  if(this->gpioEnabled == true){
    this->angle = 0.0;

    wiringPiISR(STEP, INT_EDGE_FALLING, &intPWMEdge);

    pinMode(SLEEP, OUTPUT);
    pinMode(DIR ,OUTPUT);
    pinMode(STEP, PWM_OUTPUT);
    pwmSetMode(PWM_MODE_MS);


    // モータ停止
    digitalWrite(SLEEP, 0);

    // PWM設定（初期はDuty0として停止)
    pwmSetClock(100);
    pwmSetRange(1024);
    pwmWrite(STEP, 0);
  }else{
    // GPIO初期設定エラーの場合は何もしない
  }

  return RET_SUCCESS;
}

/**
 * @brief 位置制御を行う。引数で指定した目標位置までアクチュエータを駆動する。
 * @param xRef 目標位置[cm]
 * @return -
 * @sa -
 */
void MotorController::movePosition(float xRef){
  float x = this->angle * DEG_TO_MM;
  float e;
  float pClause;
  float speedRef;

  static unsigned int judgeStopCnt;

  // error計算
  e = xRef - x;

  if((e <= 0.5) && (e >= -0.5)){
    speedRef = 0;

    if(judgeStopCnt >= 2){
      this->stopFlag = true;
    }else{
      judgeStopCnt++;
    }
  }else{

    // PI制御
    pClause = e * P_GAIN;

    iClause += e * I_GAIN;
    if(iClause >= I_CLAUSE_LIMIT){
      iClause = I_CLAUSE_LIMIT;
    }else if(iClause <= -I_CLAUSE_LIMIT){
      iClause = -I_CLAUSE_LIMIT;
    }

#if ANTI_WINDUP == 1
    if( (pClause >= 0 && iClause < 0) || (pClause < 0 && iClause >= 0)){
      iClause = 0;
    }
#endif

    speedRef = pClause + iClause;

    if((e >= 1) || ( e <= -1)){
      this->stopFlag = false;
    }
    judgeStopCnt = 0;
  }

  driveSpeed(speedRef);
}

/**
 * @brief 指定したrpsと、方向でモータを駆動
 * @param speed 回転速度[rps]（正ならCW、負ならCCW）
 * @return -
 * @sa -
 */
void MotorController::driveSpeed(float speed){
  float ref = 0.0;
  if(this->gpioEnabled == true){
    ref = speed * 10*2*3.14159*0.001;

    if(ref > MAX_SPEED){
      ref = MAX_SPEED;
    }else if(ref < -MAX_SPEED){
      ref = -MAX_SPEED;
    }

    pthread_mutex_lock(&mutex);
    // PWM設定
    if( (ref <= 0.01) && (ref >= -0.01)){
      // PWM設定を停止
      pwmSetClock(100);
      pwmSetRange(1024);
      pwmWrite(STEP, 0);

      // モータ停止
      digitalWrite(SLEEP, 0);
    }else if(ref > 0.0){
      pwmSetClock((int)(RPS_TO_CLOCK / ref));
      pwmSetRange(1024);
      pwmWrite(STEP, 512);

      // DIR
      digitalWrite(DIR, DIR_CCW);

      // SLEEP解除
      digitalWrite(SLEEP, 1);
    }else{
      pwmSetClock((int)(RPS_TO_CLOCK / -ref));
      pwmSetRange(1024);
      pwmWrite(STEP, 512);

      // DIR
      digitalWrite(DIR, DIR_CW);

      // SLEEP解除
      digitalWrite(SLEEP, 1);
    }

    pthread_mutex_unlock(&mutex);
  }
}


/**
 * @brief 現在位置を取得する
 * @param -
 * @return 現在位置[cm]
 * @sa -
 */
float MotorController::getPosition(void){
  return this->angle * (float)DEG_TO_MM;
}


/**
 * @brief 停止しているかを調べる
 * @param -
 * @return 停止フラグ(true:停止 false:停止していない）
 * @sa -
 */
bool MotorController::isStop(){
  return this->stopFlag;
}


//////////////////////////////////////////////////////////
// private
//////////////////////////////////////////////////////////
